from unittest import TestCase

from AniPy.Anilist import Anilist
from AniPy.Anilist.models import Media


class TestAnilist(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.anilist = Anilist()

    def test_simple(self):
        anime = Media(self.anilist, 279)
        self.assertEqual(anime.idMal, 279)
        self.assertEqual(anime.isAdult, False)
        self.assertEqual(anime.isFavourite, False)
        self.assertEqual(anime.isLicensed, True)

    def test_prepare_shallow(self):
        anime = Media(self.anilist, 279)
        anime.prepare()
        self.assertNotEqual(anime.idMal, 279)
        self.assertNotEqual(anime.type, "ANIME")
        self.assertNotEqual(anime.format, "TV")
        self.assertNotEqual(anime.status, "FINISHED")
        self.assertNotEqual(anime.isLicensed, True)
        self.assertNotEqual(anime.description, """Tired of writing riddles for children, Yamaoka Momosuke plans on gathering spooky and gruesome stories and publishing them in an anthology called Hyakumonogatari ("One Hundred Tales"). While researching these old myths and legends he comes across a mysterious trio who call themselves the Ongyou. They are detectives who are investigating the legends to reveal their truths...and bring those in the wrong to justice. Each time Momosuke meets the Ongyou he must face horrible truths and battle with his morals, but he's seeing things he shouldn't be seeing... <br><br>\n(Source: ANN)""")
        self.assertNotEqual(anime.season, "FALL")
        self.assertNotEqual(anime.episodes, 13)
        self.assertNotEqual(anime.duration, 24)
        self.assertNotEqual(anime.chapters, None)
        self.assertNotEqual(anime.volumes, None)
        self.assertNotEqual(anime.countryOfOrigin, "JP")
        self.assertNotEqual(anime.isLicensed, True)
        self.assertNotEqual(anime.source, None)
        self.assertNotEqual(anime.hashtag, None)
        self.assertNotEqual(anime.bannerImage, None)
        self.assertNotEqual(anime.averageScore, 65)
        self.assertNotEqual(anime.meanScore, 68)
        self.assertNotEqual(anime.trending, 0)
        self.assertNotEqual(anime.favourites, 11)
        self.assertNotEqual(anime.isFavourite, False)
        self.assertNotEqual(anime.isAdult, False)
        self.assertNotEqual(anime.siteUrl, "https://anilist.co/anime/279")
        self.assertNotEqual(anime.autoCreateForumThread, False)

        anime.fetch()

        self.assertEqual(anime.idMal, 279)
        self.assertEqual(anime.type, "ANIME")
        self.assertEqual(anime.format, "TV")
        self.assertEqual(anime.status, "FINISHED")
        self.assertEqual(anime.isLicensed, True)
        self.assertEqual(anime.description, """Tired of writing riddles for children, Yamaoka Momosuke plans on gathering spooky and gruesome stories and publishing them in an anthology called Hyakumonogatari ("One Hundred Tales"). While researching these old myths and legends he comes across a mysterious trio who call themselves the Ongyou. They are detectives who are investigating the legends to reveal their truths...and bring those in the wrong to justice. Each time Momosuke meets the Ongyou he must face horrible truths and battle with his morals, but he's seeing things he shouldn't be seeing... <br><br>\n(Source: ANN)""")
        self.assertEqual(anime.season, "FALL")
        self.assertEqual(anime.episodes, 13)
        self.assertEqual(anime.duration, 24)
        self.assertEqual(anime.chapters, None)
        self.assertEqual(anime.volumes, None)
        self.assertEqual(anime.countryOfOrigin, "JP")
        self.assertEqual(anime.isLicensed, True)
        self.assertEqual(anime.source, None)
        self.assertEqual(anime.hashtag, None)
        self.assertEqual(anime.bannerImage, None)
        self.assertEqual(anime.averageScore, 65)
        self.assertEqual(anime.meanScore, 68)
        self.assertEqual(anime.trending, 0)
        self.assertEqual(anime.favourites, 11)
        self.assertEqual(anime.isFavourite, False)
        self.assertEqual(anime.isAdult, False)
        self.assertEqual(anime.siteUrl, "https://anilist.co/anime/279")
        self.assertEqual(anime.autoCreateForumThread, False)

    def test_prepare_deep(self):
        anime = Media(self.anilist, 345)
        anime.prepare()
        self.assertNotEqual(anime.title.english, "Emma: A Victorian Romance")

        anime.fetch()
        self.assertEqual(anime.title.english, "Emma: A Victorian Romance")

    def test_start_and_end_date(self):
        anime = Media(self.anilist, 632)
        anime.prepare()
        self.assertNotEqual(anime.startDate.year, 2001)
        self.assertNotEqual(anime.endDate.day, 14)

        anime.fetch()

        self.assertEqual(anime.startDate.year, 2001)
        self.assertEqual(anime.endDate.day, 14)

    def test_cover_image(self):
        anime = Media(self.anilist, 123)
        anime.prepare()
        self.assertNotEqual(anime.coverImage.large, "https://cdn.anilist.co/img/dir/anime/reg/123.jpg")
        self.assertNotEqual(anime.coverImage.medium, "https://cdn.anilist.co/img/dir/anime/med/123.jpg")

        anime.fetch()

        self.assertEqual(anime.coverImage.large, "https://cdn.anilist.co/img/dir/anime/reg/123.jpg")
        self.assertEqual(anime.coverImage.medium, "https://cdn.anilist.co/img/dir/anime/med/123.jpg")

    def test_tags(self):
        test_data = [
            ('Shounen', 'Target demographic is teenage and young adult males.', False),
            ('Post-Apocalyptic', 'Partly or completely set in a world or civilization after a global disaster.', False),
            ('Guns', 'Prominently features the use of guns in combat.', False),
            ('Survival', 'Centers around the struggle to live in spite of extreme obstacles.', False),
        ]

        anime = Media(self.anilist, 25)
        anime.prepare()
        self.assertNotEqual(anime.tags.name, "")
        self.assertNotEqual(anime.tags.description, "")
        self.assertNotEqual(anime.tags.isAdult, "")

        anime.fetch()

        for index, test_element in enumerate(test_data):
            self.assertEqual(anime.tags[index].name, test_element[0])
            self.assertEqual(anime.tags[index].description, test_element[1])
            self.assertEqual(anime.tags[index].isAdult, test_element[2])

    def test_external_links(self):
        test_data = [
            (1552, 'http://www.funimation.com/shows/desert-punk/videos/episodes', 'Funimation'),
            (1967, 'http://www.hulu.com/desert-punk', 'Hulu'),
            (6614, 'http://www.crunchyroll.com/desert-punk', 'Crunchyroll')

        ]
        anime = Media(self.anilist, 25)
        anime.prepare()

        self.assertNotEqual(anime.externalLinks.id, 1)
        self.assertNotEqual(anime.externalLinks.url, "")
        self.assertNotEqual(anime.externalLinks.site, "")

        anime.fetch()

        for index, test_element in enumerate(test_data):
            self.assertEqual(anime.externalLinks[index].id, test_element[0])
            self.assertEqual(anime.externalLinks[index].url, test_element[1])
            self.assertEqual(anime.externalLinks[index].site, test_element[2])

    def test_streaming_episodes(self):
        test_data = [
            ('Episode 1 - Asteroid Blues', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/e3a45e86c597fe16f02d29efcadedcd81473268732_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-1-asteroid-blues-719653', 'Crunchyroll'),
            ('Episode 2 - Stray Dog Strut', 'https://img1.ak.crunchyroll.com/i/spire3-tmb/0eea0ce5c2a92b72fd633e5cbabd17b81473277532_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-2-stray-dog-strut-719655', 'Crunchyroll'),
            ('Episode 3 - Honky Tonk Women', 'https://img1.ak.crunchyroll.com/i/spire4-tmb/f007ab7cb899b5ca83b8e8d07a1067d51473125168_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-3-honky-tonk-women-719475', 'Crunchyroll'),
            ('Episode 4 - Gateway Shuffle', 'https://img1.ak.crunchyroll.com/i/spire4-tmb/8395f1edb9b86ec69777bdd84a9bf7921473115214_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-4-gateway-shuffle-719477', 'Crunchyroll'),
            ('Episode 5 - Ballad of Fallen Angels', 'https://img1.ak.crunchyroll.com/i/spire3-tmb/180d486bb9c563491f417eea1985c1de1473112114_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-5-ballad-of-fallen-angels-719479', 'Crunchyroll'),
            ('Episode 6 - Sympathy for the Devil', 'https://img1.ak.crunchyroll.com/i/spire1-tmb/24668be96b09d1356709e04a3daa161f1473114965_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-6-sympathy-for-the-devil-719481', 'Crunchyroll'),
            ('Episode 7 - Heavy Metal Queen', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/d2d71facab3a707cd6480df47c78c2fe1473112459_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-7-heavy-metal-queen-719483', 'Crunchyroll'),
            ('Episode 8 - Waltz for Venus', 'https://img1.ak.crunchyroll.com/i/spire4-tmb/c82f024aa8e91377f1fd1f31a732ed211473138815_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-8-waltz-for-venus-719485', 'Crunchyroll'),
            ('Episode 9 - Jamming with Edward', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/48d53aa4330ef9268c31db12def817ae1473113990_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-9-jamming-with-edward-719487', 'Crunchyroll'),
            ('Episode 10 - Ganymede Elegy', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/0b1fe4247b00a2a96cd724c978f2e9301473126093_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-10-ganymede-elegy-719489', 'Crunchyroll'),
            ('Episode 11 - Toys in the Attic', 'https://img1.ak.crunchyroll.com/i/spire1-tmb/77098545eaa70378a9ef1cdd995015641473117569_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-11-toys-in-the-attic-719491', 'Crunchyroll'),
            ('Episode 12 - Jupiter Jazz (part 1)', 'https://img1.ak.crunchyroll.com/i/spire3-tmb/cd71c0f3bcf0cb5559d0939e583daf361473118196_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-12-jupiter-jazz-part-1-719493', 'Crunchyroll'),
            ('Episode 13 - Jupiter Jazz (part 2)', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/e289707b5fc232b037a3e7d0e2a7959c1473119423_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-13-jupiter-jazz-part-2-719495', 'Crunchyroll'),
            ('Episode 14 - Bohemian Rhapsody', 'https://img1.ak.crunchyroll.com/i/spire3-tmb/44580eb7cc5740f63f3170265de80b561473120654_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-14-bohemian-rhapsody-719499', 'Crunchyroll'),
            ('Episode 15 - My Funny Valentine', 'https://img1.ak.crunchyroll.com/i/spire1-tmb/efb53c03f2ec8c0470275bb3cd32c6aa1473119912_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-15-my-funny-valentine-719503', 'Crunchyroll'),
            ('Episode 16 - Black Dog Serenade', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/c4138c03816a57f8e28e112875767dff1473121102_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-16-black-dog-serenade-719505', 'Crunchyroll'),
            ('Episode 17 - Mushroom Samba', 'https://img1.ak.crunchyroll.com/i/spire3-tmb/2140460e2f69b8a124f812b31b2f0e991473122552_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-17-mushroom-samba-719517', 'Crunchyroll'),
            ('Episode 18 - Speak Like a Child', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/22ab56e337990440b8b3dde252b64b321473122227_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-18-speak-like-a-child-719519', 'Crunchyroll'),
            ('Episode 19 - Wild Horses', 'https://img1.ak.crunchyroll.com/i/spire1-tmb/3288ad0b33dcfc5e1af91497d280b4fe1473124229_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-19-wild-horses-719521', 'Crunchyroll'),
            ('Episode 20 - Pierrot Le Fou', 'https://img1.ak.crunchyroll.com/i/spire3-tmb/785ef5c25a046832b3513089c6f3d9941473123758_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-20-pierrot-le-fou-719523', 'Crunchyroll'),
            ('Episode 21 - Boogie Woogie Feng Shui', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/13dbb524b4b6bbc4cd1d36719d789b2f1473126565_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-21-boogie-woogie-feng-shui-719525', 'Crunchyroll'),
            ('Episode 22 - Cowboy Funk', 'https://img1.ak.crunchyroll.com/i/spire1-tmb/c71fbb3f7b06b6747208d3d966aaa1b91473125734_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-22-cowboy-funk-719527', 'Crunchyroll'),
            ('Episode 23 - Brain Scratch', 'https://img1.ak.crunchyroll.com/i/spire3-tmb/0d7ddcae6489ee5aae1c2cccf09f73991473126176_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-23-brain-scratch-719529', 'Crunchyroll'),
            ('Episode 24 - Hard Luck Woman', 'https://img1.ak.crunchyroll.com/i/spire4-tmb/9d2308b354746a3fb84152526b195e1c1473128052_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-24-hard-luck-woman-719531', 'Crunchyroll'),
            ('Episode 25 - The Real Folk Blues (part 1)', 'https://img1.ak.crunchyroll.com/i/spire2-tmb/7e9bb2a05eb81b701cdf115710d659291473127483_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-25-the-real-folk-blues-part-1-719533', 'Crunchyroll'),
            ('Episode 26 - The Real Folk Blues (part 2)', 'https://img1.ak.crunchyroll.com/i/spire1-tmb/191b230426f0b0e6568b4ca6edab47321473136587_full.jpg', 'http://www.crunchyroll.com/cowboy-bebop/episode-26-the-real-folk-blues-part-2-719537', 'Crunchyroll')
        ]

        anime = Media(self.anilist, 1)
        anime.prepare()

        self.assertNotEqual(anime.streamingEpisodes.title, "")
        self.assertNotEqual(anime.streamingEpisodes.thumbnail, "")
        self.assertNotEqual(anime.streamingEpisodes.url, "")
        self.assertNotEqual(anime.streamingEpisodes.site, "")

        anime.fetch()

        for index, test_element in enumerate(test_data):
            self.assertEqual(anime.streamingEpisodes[index].title, test_element[0])
            self.assertEqual(anime.streamingEpisodes[index].thumbnail, test_element[1])
            self.assertEqual(anime.streamingEpisodes[index].url, test_element[2])
            self.assertEqual(anime.streamingEpisodes[index].site, test_element[3])

    def test_rankings(self):
        test_data = [
            'RATED',
            'POPULAR',
            'RATED',
            'POPULAR',
        ]

        anime = Media(self.anilist, 63)
        anime.prepare()

        self.assertNotEqual(anime.rankings.rank, "")
        self.assertNotEqual(anime.rankings.type, "")

        anime.fetch()

        for index, test_element in enumerate(test_data):
            self.assertEqual(anime.rankings[index].type, test_element)

    def test_require_all(self):
        anime = Media(self.anilist, 1)
        anime.prepare()
        anime.title.require_all()

        anime.fetch()

        self.assertEqual(anime.title.romaji, "Cowboy Bebop")
        self.assertEqual(anime.title.english, "Cowboy Bebop")
        self.assertEqual(anime.title.native, "カウボーイビバップ")
        self.assertEqual(anime.title.userPreferred, "Cowboy Bebop")

        anime.prepare()
        anime.require_all()
        anime.fetch()

    def tests_stats(self):
        anime = Media(self.anilist, 254)
        anime.prepare()

        self.assertNotEqual(anime.stats.scoreDistribution.score, 0)
        self.assertNotEqual(anime.stats.scoreDistribution.amount, 0)
        self.assertNotEqual(anime.stats.statusDistribution.status, "")
        self.assertNotEqual(anime.stats.statusDistribution.amount, 0)

        anime.fetch()

        self.assertNotEqual(len(anime.stats.scoreDistribution), 0)
        for distribution in anime.stats.scoreDistribution:
            self.assertIsInstance(distribution.score, int)
            self.assertIsInstance(distribution.amount, int)

        self.assertNotEqual(len(anime.stats.statusDistribution), 0)
        for distribution in anime.stats.statusDistribution:
            self.assertIsInstance(distribution.status, str)
            self.assertIsInstance(distribution.amount, int)

    def test_studios(self):
        test_data = [
            (True, 14, 'Sunrise', 'https://anilist.co/studio/14', False),
            (False, 23, 'Bandai Visual', 'https://anilist.co/studio/23', False),
            (False, 233, 'Bandai Entertainment', 'https://anilist.co/studio/233', False)
        ]

        anime = Media(self.anilist, 1)
        anime.prepare()
        self.assertNotEqual(anime.studios.isMain, True)
        self.assertNotEqual(anime.studios.id, 1)
        self.assertNotEqual(anime.studios.name, "")
        self.assertNotEqual(anime.studios.siteUrl, "")
        self.assertNotEqual(anime.studios.isFavourite, False)

        anime.fetch()

        for index, test_element in enumerate(test_data):
            self.assertEqual(anime.studios[index].isMain, test_element[0])
            self.assertEqual(anime.studios[index].id, test_element[1])
            self.assertEqual(anime.studios[index].name, test_element[2])
            self.assertEqual(anime.studios[index].siteUrl, test_element[3])
            self.assertEqual(anime.studios[index].isFavourite, test_element[4])

    def test_relations(self):
        test_data = [
            (30173, 'Cowboy Bebop'), (30174, 'Cowboy Bebop Shooting Star'),
            (5, "Cowboy Bebop: Knockin' on Heaven's Door"), (17205, "Ein's Summer Vacation"),
            (4037, 'Cowboy Bebop Session XX')
        ]

        anime = Media(self.anilist, 1)
        anime.prepare()
        self.assertNotEqual(anime.relations.node.title.english, "")

        anime.fetch()

        for index, test_element in enumerate(test_data):
            self.assertEqual(anime.relations[index].id, test_element[0])
            self.assertEqual(anime.relations[index].title.english, test_element[1])

    def test_search_anime(self):
        test_data = [
            (5114, 'Fullmetal Alchemist: Brotherhood'),
            (6421, 'Fullmetal Alchemist: Brotherhood OVA Collection'),
            (7902, 'Fullmetal Alchemist: Brotherhood: 4-Koma Theater'),
            (121, 'Fullmetal Alchemist'),
            (430, 'Fullmetal Alchemist: The Movie - Conqueror of Shamballa'),
            (664, 'Fullmetal Alchemist: Reflections'),
            (908, 'Fullmetal Alchemist: Premium OVA Collection'),
            (9135, 'Fullmetal Alchemist: The Sacred Star of Milos'),
            (10842, 'Fullmetal Alchemist: Milos no Seinaru Hoshi Specials')
        ]
        alchemist = self.anilist.search.anime("Fullmetal Alchemist")

        for index, test_element in enumerate(test_data):
            self.assertEqual(alchemist[index].id, test_element[0])
            self.assertEqual(alchemist[index].title.english, test_element[1])



import copy

from AniPy.base_models import RequireFetch, Resource, ResourceContainer, is_model
from AniPy.helper import auto_fetch


class Studio(ResourceContainer):
    _id = RequireFetch(int)
    _name = RequireFetch(str)
    _siteUrl = RequireFetch(str)
    _isFavourite = RequireFetch(str)
    _favourites = RequireFetch(int)
    _isMain = RequireFetch(bool)

    @auto_fetch
    def id(self):
        return self._id

    @auto_fetch
    def name(self):
        return self._name

    @auto_fetch
    def siteUrl(self):
        return self._siteUrl

    @auto_fetch
    def isFavourite(self):
        return self._isFavourite

    @auto_fetch
    def favourites(self):
        return self._favourites

    @auto_fetch
    def isMain(self):
        return self._isMain

    def fill(self, jsd, jsd_key):
        if "edges" in jsd:
            jsd = jsd["edges"]
            results = []
            for element in jsd:
                node = element.pop("node")
                element.update(node)
                instance = copy.deepcopy(self)
                instance.fill(element, jsd_key)
                results.append(instance)
            return results
        else:
            super().fill(jsd, jsd_key)

    def query(self):
        fields = self.fields()
        field_is_main = fields.pop("_isMain")

        result = "{\n"
        result += "edges {"
        if field_is_main.field is not None:
            result += field_is_main.field + "\n"
            if is_model(field_is_main.type.__name__):
                result += field_is_main.instance.query()

        result += "node {\n"

        for field in fields.values():
            if field.field is not None:
                result += field.field + "\n"
                if is_model(field.type.__name__):
                    result += field.instance.query()
        result += "}\n}\n}\n"
        return result


class ScoreDistribution(ResourceContainer):
    _score = RequireFetch(int)
    _amount = RequireFetch(int)

    @auto_fetch
    def score(self):
        return self._score

    @auto_fetch
    def amount(self):
        return self._amount


class StatusDistribution(ResourceContainer):
    _status = RequireFetch(str)
    _amount = RequireFetch(int)

    @auto_fetch
    def status(self):
        return self._status

    @auto_fetch
    def amount(self):
        return self._amount


class Stats(ResourceContainer):
    _scoreDistribution = RequireFetch(ScoreDistribution)
    _statusDistribution = RequireFetch(StatusDistribution)

    @auto_fetch
    def scoreDistribution(self):
        return self._scoreDistribution

    @auto_fetch
    def statusDistribution(self):
        return self._statusDistribution


class Ranking(ResourceContainer):
    _id = RequireFetch(int)
    _rank = RequireFetch(int)
    _type = RequireFetch(str)
    _format = RequireFetch(str)
    _year = RequireFetch(int)
    _season = RequireFetch(str)
    _allTime = RequireFetch(bool)
    _context = RequireFetch(str)

    @auto_fetch
    def id(self):
        return self._id

    @auto_fetch
    def rank(self):
        return self._rank

    @auto_fetch
    def type(self):
        return self._type

    @auto_fetch
    def format(self):
        return self._format

    @auto_fetch
    def year(self):
        return self._year

    @auto_fetch
    def season(self):
        return self._season

    @auto_fetch
    def allTime(self):
        return self._allTime

    @auto_fetch
    def context(self):
        return self._context


class StreamingEpisode(ResourceContainer):
    _title = RequireFetch(str)
    _thumbnail = RequireFetch(str)
    _url = RequireFetch(str)
    _site = RequireFetch(str)

    @auto_fetch
    def title(self):
        return self._title

    @auto_fetch
    def thumbnail(self):
        return self._thumbnail

    @auto_fetch
    def url(self):
        return self._url

    @auto_fetch
    def site(self):
        return self._site


class ExternalLink(ResourceContainer):
    _id = RequireFetch(int)
    _url = RequireFetch(str)
    _site = RequireFetch(str)

    @auto_fetch
    def id(self):
        return self._id

    @auto_fetch
    def url(self):
        return self._url

    @auto_fetch
    def site(self):
        return self._site


class Tag(ResourceContainer):
    _id = RequireFetch(int)
    _name = RequireFetch(str)
    _description = RequireFetch(str)
    _category = RequireFetch(str)
    _rank = RequireFetch(int)
    _isGeneralSpoiler = RequireFetch(bool)
    _isMediaSpoiler = RequireFetch(bool)
    _isAdult = RequireFetch(bool)

    @auto_fetch
    def id(self):
        return self._id

    @auto_fetch
    def name(self):
        return self._name

    @auto_fetch
    def description(self):
        return self._description

    @auto_fetch
    def category(self):
        return self._category

    @auto_fetch
    def rank(self):
        return self._rank

    @auto_fetch
    def isGeneralSpoiler(self):
        return self._isGeneralSpoiler

    @auto_fetch
    def isMediaSpoiler(self):
        return self._isMediaSpoiler

    @auto_fetch
    def isAdult(self):
        return self._isAdult


class CoverImage(ResourceContainer):
    _large = RequireFetch(str)
    _medium = RequireFetch(str)

    @auto_fetch
    def large(self):
        return self._large

    @auto_fetch
    def medium(self):
        return self._medium


class StartDate(ResourceContainer):
    _year = RequireFetch(int)
    _month = RequireFetch(int)
    _day = RequireFetch(int)

    @auto_fetch
    def year(self):
        return self._year

    @auto_fetch
    def month(self):
        return self._month

    @auto_fetch
    def day(self):
        return self._day


EndDate = StartDate


class Title(ResourceContainer):
    _romaji = RequireFetch(str)
    _english = RequireFetch(str)
    _native = RequireFetch(str)
    _userPreferred = RequireFetch(str)

    @auto_fetch
    def romaji(self):
        return self._romaji

    @auto_fetch
    def english(self):
        return self._english

    @auto_fetch
    def native(self):
        return self._native

    @auto_fetch
    def userPreferred(self):
        return self._userPreferred


class Media(Resource):
    def __init__(self, anilist=None, id=None):
        Resource.__init__(self, anilist=anilist, id=id)

        self._prepare = False

        self._id = id
        if id is None:
            self._id = RequireFetch(int, field="id")

        self._idMal = RequireFetch(int)
        self._type = RequireFetch(str)
        self._format = RequireFetch(str)
        self._status = RequireFetch(str)
        self._isLicensed = RequireFetch(bool)
        self._description = RequireFetch(str)
        self._season = RequireFetch(str)
        self._episodes = RequireFetch(int)
        self._duration = RequireFetch(int)
        self._chapters = RequireFetch(int)
        self._volumes = RequireFetch(int)
        self._countryOfOrigin = RequireFetch(str)
        self._isLicensed = RequireFetch(int)
        self._source = RequireFetch(str)
        self._hashtag = RequireFetch(str)
        self._updatedAt = RequireFetch(int)
        self._bannerImage = RequireFetch(str)
        self._averageScore = RequireFetch(int)
        self._meanScore = RequireFetch(int)
        self._popularity = RequireFetch(int)
        self._trending = RequireFetch(int)
        self._favourites = RequireFetch(int)
        self._isFavourite = RequireFetch(bool)
        self._isAdult = RequireFetch(bool)
        self._siteUrl = RequireFetch(str)
        self._autoCreateForumThread = RequireFetch(bool)

        self._title = RequireFetch(Title)
        self._startDate = RequireFetch(StartDate)
        self._endDate = RequireFetch(EndDate)
        self._coverImage = RequireFetch(CoverImage)

        self._tags = RequireFetch(Tag)
        self._externalLinks = RequireFetch(ExternalLink)
        self._streamingEpisodes = RequireFetch(StreamingEpisode)
        self._rankings = RequireFetch(Ranking)
        self._stats = RequireFetch(Stats)
        self._studios = RequireFetch(Studio)
        self._relations = RequireFetch(Relation)

    def fetch(self):
        self._prepare = False
        self._anilist.get.fetch(self)

    def prepare(self):
        self._prepare = True

    def _load(self, element):
        return not self._prepare

    @auto_fetch
    def id(self):
        return self._id

    @auto_fetch
    def idMal(self):
        return self._idMal

    @auto_fetch
    def type(self):
        return self._type

    @auto_fetch
    def format(self):
        return self._format

    @auto_fetch
    def status(self):
        return self._status

    @auto_fetch
    def isLicensed(self):
        return self._isLicensed

    @auto_fetch
    def description(self):
        return self._description

    @auto_fetch
    def season(self):
        return self._season

    @auto_fetch
    def episodes(self):
        return self._episodes

    @auto_fetch
    def duration(self):
        return self._duration

    @auto_fetch
    def chapters(self):
        return self._chapters

    @auto_fetch
    def volumes(self):
        return self._volumes

    @auto_fetch
    def countryOfOrigin(self):
        return self._countryOfOrigin

    @auto_fetch
    def isLicensed(self):
        return self._isLicensed

    @auto_fetch
    def source(self):
        return self._source

    @auto_fetch
    def hashtag(self):
        return self._hashtag

    @auto_fetch
    def updatedAt(self):
        return self._updatedAt

    @auto_fetch
    def bannerImage(self):
        return self._bannerImage

    @auto_fetch
    def averageScore(self):
        return self._averageScore

    @auto_fetch
    def meanScore(self):
        return self._meanScore

    @auto_fetch
    def popularity(self):
        return self._popularity

    @auto_fetch
    def trending(self):
        return self._trending

    @auto_fetch
    def favourites(self):
        return self._favourites

    @auto_fetch
    def isFavourite(self):
        return self._isFavourite

    @auto_fetch
    def isAdult(self):
        return self._isAdult

    @auto_fetch
    def siteUrl(self):
        return self._siteUrl

    @auto_fetch
    def autoCreateForumThread(self):
        return self._autoCreateForumThread

    @auto_fetch
    def title(self):
        return self._title

    @auto_fetch
    def startDate(self):
        return self._startDate

    @auto_fetch
    def endDate(self):
        return self._endDate

    @auto_fetch
    def coverImage(self):
        return self._coverImage

    @auto_fetch
    def tags(self):
        return self._tags

    @auto_fetch
    def externalLinks(self):
        return self._externalLinks

    @auto_fetch
    def streamingEpisodes(self):
        return self._streamingEpisodes

    @auto_fetch
    def rankings(self):
        return self._rankings

    @auto_fetch
    def stats(self):
        return self._stats

    @auto_fetch
    def studios(self):
        return self._studios

    @auto_fetch
    def relations(self):
        return self._relations


class Relation(ResourceContainer):
    _relationType = RequireFetch(str)
    _node = RequireFetch(Media)

    @auto_fetch
    def relationType(self):
        return self._relationType

    @auto_fetch
    def node(self):
        return self._node

    def query(self):
        fields = self.fields()
        field_relation_type = fields.pop("_relationType")

        result = "{\n"
        result += "edges {"
        if field_relation_type.field is not None:
            result += field_relation_type.field + "\n"
            if is_model(field_relation_type.type.__name__):
                result += field_relation_type.instance.query()

        for field in fields.values():
            if field.field is not None:
                result += field.field + "\n"
                if is_model(field.type.__name__):
                    result += field.instance.query()
        result += "}\n}\n"
        return result

    def fill(self, jsd, jsd_key):
        if "edges" in jsd:
            jsd = jsd["edges"]
            results = []
            for element in jsd:
                node = element.pop("node")
                instance = copy.deepcopy(self.node)
                instance.fill(node, jsd_key)
                results.append(instance)
            return results
        else:
            super().fill(jsd, jsd_key)



import copy
import sys


class ResourceContainer:
    def __init__(self, anilist, *args, **kwargs):
        self._anilist = anilist

    def require_all(self, recursive=False):
        for key in self.fields():
            if not isinstance(getattr(self, key), RequireFetch):
                continue

            if is_model(getattr(self, key).type.__name__):
                if recursive:
                    getattr(self, key[1:]).require_all(recursive)

                continue

            getattr(self, key[1:])

    def _load(self, element):
        return False

    def fields(self):
        return retrieve_required_fields(self)

    def fill(self, jsd, jsd_key):
        if isinstance(jsd, dict):
            for key in jsd:
                getattr(self, "_" + key).fill(jsd[key], key, parent=self)
        if isinstance(jsd, list):
            results = []
            for element in jsd:
                instance = copy.deepcopy(self)
                instance.fill(element, jsd_key)
                results.append(instance)
            return results

    def query(self):
        fields = self.fields()
        result = "{\n"
        for field in fields.values():
            if field.field is not None:
                result += field.field + "\n"
                if is_model(field.type.__name__):
                    result += field.instance.query()
        result += "}\n"
        return result


class Resource(ResourceContainer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def fetch(self):
        raise NotImplementedError()

    def _load(self, element):
        return True


class RequireFetch:
    def __init__(self, type=None, field=None):
        self.type = type
        self.field = field
        self.instance = None
        if field is not None:
            self.instance = type()

    def fill(self, jsd, jsd_key, parent=None):
        if is_model(self.type.__name__):
            if self.instance is None:
                if is_model(self.type.__name__):
                    self.instance = self.type(parent._anilist)
                else:
                    self.instance = self.type()
            result = self.instance.fill(jsd, jsd_key)
            if result is not None:
                self.instance = result
            setattr(parent, "_" + (self.field if self.field is not None else jsd_key), self.instance)
        else:
            setattr(parent, "_" + (self.field if self.field is not None else jsd_key), jsd)


def is_model(type):
    type = type[0].upper() + type[1:]
    return getattr(sys.modules["AniPy.Anilist.models"], type, False)


def retrieve_required_fields(model):
    fields = [a for a in model.__dir__() if a.startswith("_") and not a.startswith("__")]

    result = {}

    for field in fields:
        field_data = getattr(model, field)
        if isinstance(field_data, RequireFetch):
            result[field] = field_data

    return result



from AniPy.Anilist.get import Get
from AniPy.Anilist.search import Search


class Anilist:

    def __init__(self, cid=None, csecret=None, credentials=None):
        self.settings = {
            "header": {
                "Content-Type": "application/json",
                "User-Agent": "ani.py (git.widmer.me/project/multimedia/ani.py)",
                "Accept": "application/json"
            },
            'authurl': 'https://anilist.co/api',
            'apiurl': 'https://graphql.anilist.co',
            'cid': cid,
            'csecret': csecret,
            'token': credentials
        }

        self.search = Search(self.settings, self)
        self.get = Get(self.settings)



import json

import requests

from AniPy.Anilist.models import Media


class Search:
    def __init__(self, settings, anilist):
        self.settings = settings
        self.anilist = anilist

    def anime(self, term, page=1, perpage=25):
        query_string = """\
        query ($query: String, $page: Int, $perpage: Int) {
            Page (page: $page, perPage: $perpage) {
                pageInfo {
                    total
                    currentPage
                    lastPage
                    hasNextPage
                }
                media (search: $query, type: ANIME) {
                    id
                    title {
                        romaji
                        english
                    }
                    coverImage {
                        large
                    }
                    averageScore
                    popularity
                    episodes
                    season
                    hashtag
                    isAdult
                }
            }
        }
        """
        vars = {"query": term, "page": page, "perpage": perpage}
        r = requests.post(self.settings['apiurl'],
                          headers=self.settings['header'],
                          json={'query': query_string, 'variables': vars})
        jsd = r.text

        try:
            jsd = json.loads(jsd)
        except ValueError:
            return None
        else:
            media = jsd["data"]["Page"]["media"]
            result = []
            for element in media:
                media_element = Media(anilist=self.anilist)
                media_element.fill(element, "")
                result.append(media_element)
            return result



import sys

from AniPy.base_models import RequireFetch


def auto_fetch(func):
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)

        if not isinstance(result, RequireFetch):
            return result

        result.field = func.__name__
        if not self._load(result):

            if getattr(sys.modules["AniPy.Anilist.models"], result.type.__name__, False):
                if result.instance is None:
                    result.instance = result.type(self._anilist)

                    prepare = getattr(result.instance, "prepare", False)
                    if prepare:
                        prepare()

                    print("")
                return result.instance

            else:

                return result.type

        self.fetch()
        result = func(self, *args, **kwargs)
        if isinstance(result, RequireFetch):
            raise NotImplementedError("The fild " + str(func) + " is not included in fetch()")

        return result
    return property(wrapper)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ani.py",
    version="0.1.1",
    author="Joel Widmer",
    author_email="joel.widmer.wj@gmail.com",
    description="Python interface for anime APIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
        "requests",
        "requests-toolbelt",
    ],
)



import json
import time

import requests

from AniPy.errors import AniPyImplementationError


class Get:
    def __init__(self, settings):
        self.settings = settings
        self.session = requests.session()

        self.remaining = 90
        self.limit = 90
        self.reset = time.time()

    def fetch(self, element):
        request = element.query()

        request = """\
        query($id: Int) {
        """ + type(element).__name__ + """(id: $id, type: ANIME)
        """ + request + "}"

        vars = {"id": element.id}

        if self.remaining == 0:
            time.sleep(max(self.reset - time.time(), 0))

        r = self.session.post(self.settings["apiurl"],
                              headers=self.settings["header"],
                              json={"query": request, "variables": vars})

        jsd = json.loads(r.text)

        self.limit = int(r.headers.get("x-ratelimit-limit"))
        self.remaining = int(r.headers.get("x-ratelimit-remaining"))
        self.reset = int(r.headers.get("X-RateLimit-Reset"))

        print(self.remaining, "/", self.limit, ",", self.reset, time.time())

        if len(jsd.get("errors", [])) > 0:
            print(request)
            raise AniPyImplementationError(str(jsd["errors"][0]))

        return element.fill(jsd["data"][type(element).__name__], "")
